%A MATLAB Toolbox
%
%Compression-based Texture Merging (CTM)
%
%Authors: John Wright, and Allen Y. Yang
%Contact: Allen Y. Yang <yang@eecs.berkeley.edu>
%
%(c) Copyright. University of California, Berkeley. 2007.
%
%Notice: The packages should NOT be used for any commercial purposes
%without direct consent of their author(s). The authors are not responsible
%for any potential property loss or damage caused directly or indirectly by
%the usage of the software.

function [sampleLabels, syntheticImage, epsilon]=CTM_adaptive_segmentor(image,varargin);
%A segmentation algorithm of images based on lossy data compression
%
%References: Unsupervised segmentation of natural images via lossy data compression (preprint)
%
%Input parameters:
%   image:                      Source image to segment.
%   epsilonList:                An array of epsilon candidates. The optimal epsilon is selected based on 
%                               the threshold textureDifferenceThreshold.
%   GaussianVariance:           Variance of the Gaussian smoothing mask.
%   mergeSmallRegions:          A boolean value to indicate whether to locally merge small regions in the 
%                               postprocessing step.
%   avoidBoundaryPixels:        A boolean value to indicate whether to omit
%                               boundary texture vectors.
%   textureDifferenceMetric:    Choose betwen two texture similarity
%                               measures: mean distance or Mallows distance.

DEBUG = 0;
addpath superpixels
addpath superpixels/yu_imncut
addpath superpixels/lib/matlab

TEXTURE_MEAN_METRIC = 1;
TEXTURE_MALLOWS_METRIC = 2;

epsilonList = [0.01 0.02 0.04 0.06 0.08 0.1 0.12 0.14 0.16 0.18 0.2 0.22 0.24 0.26 0.28 0.3 0.35 0.4 0.45 0.5];
downsampleDimension = 8;
windowSize = 7;
superpixelImageFilename = [];   % Set an unassigned value
initialSampleLabels = [];       % Set an unassigned value
readSuperpixelFile = true;
readSampleLabels = false;
saveSegmentationResult = false;
colorMetric = 'lab';
channelCount = 3;
GaussianVariance = 0.5;
avoidBoundaryPixels = false;
mergeSmallRegions = true;
smallRegionThreshold = 200;
textureDifferenceMetric = TEXTURE_MEAN_METRIC;
if textureDifferenceMetric == TEXTURE_MALLOWS_METRIC
    textureDifferenceThreshold=0.4;   
else
    textureDifferenceThreshold=0.2;   
end

% Number of superpixels coarse/fine.
N_sp=200;

% Parse the optional inputs.
if (mod(length(varargin), 2) ~= 0 )
    error(['Extra Parameters passed to the function ''' mfilename ''' must be passed in pairs.']);
end
parameterCount = length(varargin)/2;

for parameterIndex = 1:parameterCount,
    parameterName = varargin{parameterIndex*2 - 1};
    parameterValue = varargin{parameterIndex*2};
    switch lower(parameterName)
        case 'epsilonlist'
            epsilonList = parameterValue;
        case 'debug'
            DEBUG = parameterValue;
        case 'texturedifferencemetric'
            textureDifferenceMetric = parameterValue;
            if parameterValue~=TEXTURE_MALLOWS_METRIC & parameterValue~=TEXTURE_MEAN_METRIC
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'texturedifferencethreshold'
            textureDifferenceThreshold = parameterValue;
            if parameterValue<=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'mergesmallregions'
            mergeSmallRegions = parameterValue;
            if parameterValue~=true & parameterValue~=false
                error(['The parameter ''' parameterName ''' must be boolean.']);
            end  
        case 'avoidboundarypixels'
            avoidBoundaryPixels = parameterValue;
            if parameterValue~=true & parameterValue~=false
                error(['The parameter ''' parameterName ''' must be boolean.']);
            end
        case 'downsampledimension'
            downsampleDimension = parameterValue;
            if parameterValue<=1
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'windowsize'
            windowSize = parameterValue;
            if parameterValue<1 || rem(parameterValue,2)==0
                error('windowSize must be an odd natural number.');
            end
        case 'superpixelimagefilename'
            superpixelImageFilename = parameterValue;
        case 'readsuperpixelfile'
            readSuperpixelFile = parameterValue;
            if parameterValue~=0 && parameterValue~=1
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'readsamplelabels'
            readSampleLabels = parameterValue;
            if parameterValue~=0 && parameterValue~=1
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'colormetric'
            colorMetric = lower(parameterValue);
            if ~strcmp(parameterValue,'rgp') && ~strcmp(parameterValue,'lab') && ...
                    ~strcmp(parameterValue,'uvl')
                error('colorMetric only supports ''RGB'', ''LAB'', and ''UVL''. ');
            end
        case 'channelcount'
            channelCount = parameterValue;
        case 'superpixelcount'
            N_sp = parameterValue;
        case 'savesegmentationresult'
            saveSegmentationResult = parameterValue;
            if parameterValue~=0 && parameterValue~=1
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'gaussianvariance'
            % The variance of the Gaussian smoothing kernel w.r.t the
            % window size.
            GaussianVariance = parameterValue;
            if parameterValue<=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        otherwise
            error([parameterName ' is an invalid parameter.']);
    end
end
if (readSuperpixelFile==true) && (isempty(superpixelImageFilename))
    error('Superpixel image file name is not defined.');
end
if (saveSegmentationResult==true) && (isempty(superpixelImageFilename))
    error('Data file name is not defined to save the segmentation result.');
end
if (windowSize<1) || (floor(windowSize/2)*2==windowSize)
    error('The parameter windowSize must be an odd natural number.');
end
% Set thresholds
if textureDifferenceThreshold==0
    if textureDifferenceMetric == TEXTURE_MEAN_METRIC
        % metric by the mean of the Gaussians
        textureDifferenceThreshold = 0.03;
    else
        % metric by the Mallows distance
        textureDifferenceThreshold = 0.003;
    end
end
load normalization.mat meanX normalizedMagnitude

if DEBUG>2
    figure(1);
    subplot(1,3,1);
    imshow(image);
end

% Obtain an initial segmentation
halfWindowSize= floor(windowSize/2);

imageDouble = im2double(image);
[imageX, imageY, imageDepth] = size(image);
if imageDepth~=3
    error('The input image must have 3 channels.');
end

if isempty(initialSampleLabels)
    if readSuperpixelFile == true

        % Test if the file exists
        filePathIndices=find(superpixelImageFilename=='/');
        filePath = superpixelImageFilename(1:filePathIndices(end));
        fileName = superpixelImageFilename(filePathIndices(end)+1:end);
        directoryFiles = dir(filePath);

        % Search for fileName
        readSuperpixelFile = false;
        for fileIndex=1:length(directoryFiles)
            if strcmp(directoryFiles(fileIndex).name,fileName)
                % The data file exists
                readSuperpixelFile = true;
                disp('loading from the mat file...');
                load(superpixelImageFilename,'superpixelLabels');
                if readSampleLabels
                    warning off
                    load(superpixelImageFilename,'sampleLabels');
                    warning on
                    if ~isempty(whos('sampleLabels'))
                        if nargout>=2
                            syntheticImage = generate_synthetic_image_segments(image,...
                                sampleLabels);
                        end

                        if DEBUG>1
                            subplot(1,3,2);
                            superpixelImage = segImage(imageDouble,superpixelLabels);
                            imshow(superpixelImage);
                            subplot(1,3,3);
                            if nargout<2
                                syntheticImage = generate_synthetic_image_segments(image,...
                                    sampleLabels);
                            end
                            imshow(syntheticImage);
                        end
                        return;
                    end
                end

                break;
            end
        end
    end

    if readSuperpixelFile ==false
        % Pre-processing step, compute superpixels

        % Number of eigenvectors.
        N_ev=40;

        % ncut parameters for superpixel computation
        diag_length = sqrt(imageX*imageX + imageY*imageY);
        par = imncut_sp;
        par.int=0;
        par.pb_ic=1;
        par.sig_pb_ic=0.05;
        par.sig_p=ceil(diag_length/50);
        par.verbose=0;
        par.nb_r=ceil(diag_length/60);
        par.rep = -0.005;  % stability?  or proximity?
        par.sample_rate=0.2;
        par.nv = N_ev;
        par.sp = N_sp;

        % Intervening contour using mfm-pb
        fprintf('running PB\n');
        [emag,ephase] = pbWrapper(imageDouble,par.pb_timing);
        emag = pbThicken(emag);
        par.pb_emag = emag;
        par.pb_ephase = ephase;
        clear emag ephase N_ev N_sp;

        fprintf('Ncutting...\n');
        superpixelLabels = imncut_sp(imageDouble,par);

        if ~isempty(superpixelImageFilename)
            save(superpixelImageFilename,'superpixelLabels');
        end

        clear par
    end
else
    superpixelLabels = initialSampleLabels;
    clear initialSampleLabels
end

if DEBUG>2
    subplot(1,3,2);
    superpixelImage = segImage(imageDouble,superpixelLabels);
    imshow(superpixelImage);
end

% crop out boundary
imageX = (imageX-2*halfWindowSize);
imageY = (imageY-2*halfWindowSize);

if size(superpixelLabels,1)==size(image,1)
    % If the superpixelLabels is an uncropped version, crop the boundary.
    superpixelLabels = superpixelLabels(1+halfWindowSize:end-halfWindowSize, 1+halfWindowSize:end-halfWindowSize);
end

% Run test on superpixelLabels
segmentCount = max(max(superpixelLabels));
segmentIndex2 = 0;
segmentSeeds = zeros(segmentCount,1);
for segmentIndex=1:segmentCount
    pixelCoordinates = find(superpixelLabels==segmentIndex);
    
    % Eliminate empty classes
    if ~isempty(pixelCoordinates)
        segmentIndex2  = segmentIndex2 + 1;
        superpixelLabels(pixelCoordinates) = segmentIndex2;
               
        segmentSeeds(segmentIndex2) = pixelCoordinates(1);
    end
end
segmentSeeds = segmentSeeds(1:segmentIndex2);
segmentCount = segmentIndex2;

completeSegmentPixelCells = cell(1,segmentCount);
for segmentIndex=1:segmentCount
    % completeSegmentPixelCells is a complete registration of the group pixels.
    completeSegmentPixelCells{segmentIndex} = find(superpixelLabels==segmentIndex)';
end

% Construct vicinityMap
vicinityMap=vicinity_map(superpixelLabels);
clear pixelCoordinates segmentIndex2

disp('****** Step I: Construct sample vectors ******');
% Construct and homogenize sample vectors
warning off
if strcmp(colorMetric,'rgb')
    imageDouble = im2double(image);
    channelCount = 3;
elseif strcmp(colorMetric,'lab')
    colorForm = makecform('srgb2lab');
    imageDouble = double(applycform(image,colorForm));
    imageDouble = imageDouble(:,:,1:channelCount);
elseif strcmp(colorMetric,'uvl')
    colorForm = makecform('srgb2xyz');
    imageDouble = applycform(image,colorForm);
    colorForm = makecform('xyz2uvl');
    imageDouble = applycform(imageDouble,colorForm);
    imageDouble = imageDouble(:,:,1:channelCount); 
end
warning on

GaussianMask = torr_gauss_mask(halfWindowSize, halfWindowSize*GaussianVariance);
sampleDimension = windowSize*windowSize*channelCount;
sampleCount = imageX*imageY;
X = zeros(sampleDimension,sampleCount);
XLabels = zeros(1,sampleCount);
maxMagnitude = 0;
sampleCount = 0;
% Construct sample vectors
for segmentIndex=1:segmentCount
    
    isAllBoundaryPixels = true;
    for index=1:length(completeSegmentPixelCells{segmentIndex})
        pixelIndex = completeSegmentPixelCells{segmentIndex}(index);
        offsetY = floor((pixelIndex-1)/imageX)+1;
        offsetX = pixelIndex-(offsetY-1)*imageX;

        windowX1=max(offsetX-halfWindowSize,1);
        windowX2=min(offsetX+halfWindowSize,imageX);
        windowY1=max(offsetY-halfWindowSize,1);
        windowY2=min(offsetY+halfWindowSize,imageY);
        labelWindow = superpixelLabels(windowX1:windowX2, windowY1:windowY2);
        
        if avoidBoundaryPixels & ~isempty(find(labelWindow~=labelWindow(1)))
                % This is a boundary pixel
                continue;
        end
        % This is a interior pixel or avoidBoundaryPixels==false
        isAllBoundaryPixels = false;
        
        sampleWindow = imageDouble(offsetX:offsetX+2*halfWindowSize, offsetY:offsetY+2*halfWindowSize,:);
        for channelIndex=1:channelCount
            sampleWindow(:,:,channelIndex) = conv2(sampleWindow(:,:,channelIndex),GaussianMask ,'same');
        end

        sampleCount = sampleCount + 1;
        X(:,sampleCount) = (reshape(sampleWindow,sampleDimension,1)-meanX)/normalizedMagnitude;
        XLabels(sampleCount) = superpixelLabels(pixelIndex);
    end
    
    if isAllBoundaryPixels
        % If no interior pixels, we are forced to use boundary pixels
        for index=1:length(completeSegmentPixelCells{segmentIndex})
            pixelIndex = completeSegmentPixelCells{segmentIndex}(index);
            offsetY = floor((pixelIndex-1)/imageX)+1;
            offsetX = pixelIndex-(offsetY-1)*imageX;

            sampleWindow = imageDouble(offsetX:offsetX+2*halfWindowSize, offsetY:offsetY+2*halfWindowSize,:);
            for channelIndex=1:channelCount
                sampleWindow(:,:,channelIndex) = conv2(sampleWindow(:,:,channelIndex),GaussianMask ,'same');
            end
            sampleCount = sampleCount + 1;
            X(:,sampleCount) = (reshape(sampleWindow,sampleDimension,1)-meanX)/normalizedMagnitude;
            XLabels(sampleCount) = superpixelLabels(pixelIndex);
        end
    end
end

segmentPixelCells = cell(1,segmentCount);
for segmentIndex=1:segmentCount
    segmentPixelCells{segmentIndex} = find(XLabels==segmentIndex);
end

clear imageDouble sampleWindow colorForm superpixelLabels offsetX offsetY boundaryMap XLabels

% downsample
if downsampleDimension<sampleDimension
    [ignored, S, V] = svds(X, downsampleDimension);
    X = S*V.';
    sampleDimension = downsampleDimension;
elseif downsampleDimension>sampleDimension
    warning('The sample dimension is less than the downsample dimension.');
end

% Segmentation
disp('****** Step II: Segmentation ******');
segmentRegistry = segmentCount;
segmentMDL = zeros(1,segmentRegistry);
epsilonIndex = 0;
previousSegmentCount=0;
while (epsilonIndex<length(epsilonList))
    epsilonIndex = epsilonIndex + 1;
    epsilon = epsilonList(epsilonIndex);
        
    if DEBUG>0
        disp(['Current epsilon: ' num2str(epsilon)]);
    end

    mergeCount = 0;
    for segmentIndex=1:segmentRegistry
        if isempty(segmentPixelCells{segmentIndex})
            segmentMDL(segmentIndex) = 0;
        else
            segmentMDL(segmentIndex) = coding_length(X(:,segmentPixelCells{segmentIndex}),sampleCount,epsilon,true);
        end
    end
    mergeCount = 0;
    mergeMDLDifference = [];
    mergeRecord = [];
    for segmentIndex=1:segmentRegistry
        if segmentMDL(segmentIndex)~=0
            currentSubset = X(:,segmentPixelCells{segmentIndex});

            for vicinityIndex=1:length(vicinityMap{segmentIndex})
                vicinitySegment = vicinityMap{segmentIndex}(vicinityIndex);
                if vicinitySegment<segmentIndex
                    % To avoid computing merging twice, enforce in mergeRecord, 1<2
                    continue;
                end
                mergeCount = mergeCount + 1;
                mergeMDLDifference(mergeCount) = segmentMDL(vicinitySegment)+segmentMDL(segmentIndex) - ...
                    coding_length([currentSubset X(:,segmentPixelCells{vicinitySegment})],sampleCount,epsilon,true);
                mergeRecord(mergeCount,1:2) = [segmentIndex vicinitySegment];
            end
        end
    end
    
    while (segmentCount>1)

        [maxMergeMDLDifference, maxMDLIndex ]= max(mergeMDLDifference);

        if maxMergeMDLDifference>0
            % Merge mergeRecord(1,1) to mergeRecord(1,2)
            mergedSegment1 = mergeRecord(maxMDLIndex,1);
            mergedSegment2 = mergeRecord(maxMDLIndex,2);

            % Update MDL
            segmentMDL(mergedSegment1) = segmentMDL(mergedSegment1) + segmentMDL(mergedSegment2)- mergeMDLDifference(maxMDLIndex);
            segmentMDL(mergedSegment2) = 0;

            % Update pixel records
            segmentPixelCells{mergedSegment1} = [segmentPixelCells{mergedSegment1} segmentPixelCells{mergedSegment2}];
            segmentPixelCells{mergedSegment2} = [];
            completeSegmentPixelCells{mergedSegment1} = [completeSegmentPixelCells{mergedSegment1} completeSegmentPixelCells{mergedSegment2}];
            completeSegmentPixelCells{mergedSegment2} = [];
            currentX = X(:,segmentPixelCells{mergedSegment1});

            % Update vicinityMap
            for vicinityIndex=1:length(vicinityMap{mergedSegment2})
                % Change all vicinityMap indices that pointing at 2 toward 1
                neighborSegment = vicinityMap{mergedSegment2}(vicinityIndex);
                vicinityMap{neighborSegment} = union(vicinityMap{neighborSegment},mergedSegment1);
                vicinityMap{neighborSegment} = setdiff(vicinityMap{neighborSegment},mergedSegment2);
            end
            vicinityMap{mergedSegment1} = setdiff(union(vicinityMap{mergedSegment1}, vicinityMap{mergedSegment2}),...
                [mergedSegment1 mergedSegment2]);
            vicinityMap{mergedSegment2} = [];

            % Update mergeMDLDifference
            mergeRecord = mergeRecord([1:maxMDLIndex-1, maxMDLIndex+1:end],:);
            mergeMDLDifference = mergeMDLDifference([1:maxMDLIndex-1, maxMDLIndex+1:end]);
            mergeCount = mergeCount - 1;

            remainingIndices = [];
            for mergeIndex=1:mergeCount
                if (mergeRecord(mergeIndex,1)~=mergedSegment1) & (mergeRecord(mergeIndex,1)~=mergedSegment2) & ...
                        (mergeRecord(mergeIndex,2)~=mergedSegment1) & (mergeRecord(mergeIndex,2)~=mergedSegment2)
                    % Keep this index item
                    remainingIndices = [remainingIndices mergeIndex];
                end
            end
            mergeRecord = mergeRecord(remainingIndices,:);
            mergeMDLDifference = mergeMDLDifference(remainingIndices);
            mergeCount = length(remainingIndices);

            % The just deleted mergeMDL need to be recomputed
            currentVicinityMap = vicinityMap{mergedSegment1};
            for mergeIndex=1:length(currentVicinityMap)
                mergeCount = mergeCount + 1;
                vicinitySegment = currentVicinityMap(mergeIndex);
                mergeRecord(mergeCount,1)=min(mergedSegment1,vicinitySegment);
                mergeRecord(mergeCount,2)=max(mergedSegment1,vicinitySegment);
                mergeMDLDifference(mergeCount) = segmentMDL(mergedSegment1)+segmentMDL(vicinitySegment) ...
                    - coding_length([currentX X(:,segmentPixelCells{vicinitySegment})],sampleCount,epsilon,true);
            end

            segmentCount = segmentCount - 1;

        else
            % Unable to further merge segments, stop the search
            break;
        end

    end
    
    % Test the texture threshold
    stopFlag = true;
    
    if textureDifferenceMetric == TEXTURE_MEAN_METRIC
        % Only compare Gaussian mean distances
        for segmentIndex=1:segmentRegistry
            if ~isempty(completeSegmentPixelCells{segmentIndex})
                segmentMean = mean(X(:,segmentPixelCells{segmentIndex})');
                neighborList = vicinityMap{segmentIndex};

                for neighborIndex=1:length(neighborList)
                    mergeIndex = neighborList(neighborIndex);

                    if norm(segmentMean-mean(X(:,segmentPixelCells{mergeIndex})'))<textureDifferenceThreshold
                        % A neighboring pair below threshold, need to merge more
                        stopFlag = false;
                        break;
                    end
                end

                if ~stopFlag
                    break;
                end
            end
        end
    else
        % Compute and compare the Mallows distance
        for segmentIndex=1:segmentRegistry
            if ~isempty(completeSegmentPixelCells{segmentIndex})
                segmentMean = mean(X(:,segmentPixelCells{segmentIndex})')';
                segmentVariance = cov(X(:,segmentPixelCells{segmentIndex})');
                neighborList = vicinityMap{segmentIndex};

                for neighborIndex=1:length(neighborList)
                    mergeIndex = neighborList(neighborIndex)';
                    neighborMeanDifference = segmentMean-mean(X(:,segmentPixelCells{mergeIndex})')';
                    neighborVariance = cov(X(:,segmentPixelCells{mergeIndex})');
                    
                    [U, S, V] = svd(segmentVariance*neighborVariance);
                    MallowsDistance = sqrt(neighborMeanDifference.'*neighborMeanDifference + ...
                        trace(segmentVariance + neighborVariance - 2*(U*sqrt(S)*V.')));
                    
                    if MallowsDistance<textureDifferenceThreshold
                        % A neighboring pair below threshold, need to merge more
                        stopFlag = false;
                        break;
                    end
                end

                if ~stopFlag
                    break;
                end
            end
        end        
    end
    
    if stopFlag
        break;
    end
end

% Create sampleLabels
disp('****** Post-processing ******');
if mergeSmallRegions
   
    smallRegionList = [];
    segmentMean = zeros(sampleDimension,segmentRegistry);
    for segmentIndex=1:segmentRegistry
        if ~isempty(completeSegmentPixelCells{segmentIndex})
            % First, compute mean
            currentX = X(:,segmentPixelCells{segmentIndex});
            segmentMean(:,segmentIndex) = mean(currentX')';

            % Second, test if it is a small region
            if length(completeSegmentPixelCells{segmentIndex})<=smallRegionThreshold
                smallRegionList = [smallRegionList segmentIndex];
            end
        end
    end
    
    smallIndex=1;
    while smallIndex<=length(smallRegionList)
        % Find the most similar neighbor
        smallRegion = smallRegionList(smallIndex);
        minDistance = inf;
        closestNeighbor = 0;
        for vicinityIndex=1:length(vicinityMap{smallRegion})
            neighborSegment = vicinityMap{smallRegion}(vicinityIndex);
            distance = norm(segmentMean(smallRegion)-segmentMean(neighborSegment));
            if distance<minDistance
                minDistance = distance;
                closestNeighbor = neighborSegment;
            end
        end
        
        % Merge
        % 1. Update segment means
        segmentSize1=length(segmentPixelCells{closestNeighbor});
        segmentSize2=length(segmentPixelCells{smallRegion});
        segmentMean(:,closestNeighbor) = ...
            (segmentSize1*segmentMean(:,closestNeighbor)+segmentSize2*segmentMean(:,smallRegion))/(segmentSize1 + segmentSize2);
        
        % 2. Update segment pixels
        completeSegmentPixelCells{closestNeighbor} = [completeSegmentPixelCells{closestNeighbor} completeSegmentPixelCells{smallRegion}];
        completeSegmentPixelCells{smallRegion}=[];
        segmentPixelCells{closestNeighbor} = [segmentPixelCells{closestNeighbor} segmentPixelCells{smallRegion}];
        segmentPixelCells{smallRegion}=[];
        
        % 3. Update vicinityMap
        vicinityMap{closestNeighbor} = union(vicinityMap{closestNeighbor}, vicinityMap{smallRegion});
        vicinityMap{closestNeighbor} = setdiff(vicinityMap{closestNeighbor}, [smallRegion closestNeighbor]);
        vicinityMap{smallRegion} = [];
        for vicinityIndex=1:length(vicinityMap{closestNeighbor})
            % Update vicinityMap for the neighbors
            neighborSegment = vicinityMap{closestNeighbor}(vicinityIndex);
            vicinityMap{neighborSegment} = setdiff(union(vicinityMap{neighborSegment},closestNeighbor),smallRegion);
        end
        
        if sum(ismember(smallRegionList, closestNeighbor))
            % two small regions were merged
            if length(completeSegmentPixelCells{closestNeighbor})>smallRegionThreshold
                smallRegionList = setdiff(smallRegionList, closestNeighbor);                
            end
        end
        
        smallIndex = smallIndex + 1;
    end
end

sampleLabels = zeros(imageX, imageY);
segmentCount = 0;
for segmentIndex=1:segmentRegistry
    if ~isempty(completeSegmentPixelCells{segmentIndex})
        segmentCount = segmentCount + 1;
        sampleLabels(completeSegmentPixelCells{segmentIndex}) = segmentCount;
    end
end

if saveSegmentationResult
    disp('Segmentation result in the data file has been overwritten.');
    save(superpixelImageFilename,'sampleLabels','-append');
end

% Update syntheticImage
if nargout>=2
    
    halfWindowSize= floor(windowSize/2);
    syntheticImage = generate_synthetic_image_segments(image(1+halfWindowSize:end-halfWindowSize,1+halfWindowSize:end-halfWindowSize,:),...
        sampleLabels);
end

if DEBUG>1
    subplot(1,3,3);
    imshow(syntheticImage);
end