function segments = retrieve_segment_samples(image, pixelLabels, downsampleDimension)
% Output:
%       segments: a K-by-2 cell structure. segments{i,1} contains interior
%       samples; segment{i,2} contains boundary samples.

if nargin<3
    downsampleDimension = 8;
end

colorMetric = 'lab';
channelCount = 3;
GaussianVariance = 0.5;

% Crop the image
imageDouble = im2double(image);
[imageX, imageY, imageDepth] = size(image);
labelX = size(pixelLabels,1);
halfWindowSize = (imageX-labelX)/2;
windowSize = halfWindowSize * 2 + 1;
imageX = (imageX-2*halfWindowSize);
imageY = (imageY-2*halfWindowSize);

% Convert color metric
warning off
if strcmp(colorMetric,'rgb')
    imageDouble = im2double(image);
    channelCount = 3;
elseif strcmp(colorMetric,'lab')
    colorForm = makecform('srgb2lab');
    imageDouble = double(applycform(image,colorForm));
    imageDouble = imageDouble(:,:,1:channelCount);
elseif strcmp(colorMetric,'uvl')
    colorForm = makecform('srgb2xyz');
    imageDouble = applycform(image,colorForm);
    colorForm = makecform('xyz2uvl');
    imageDouble = applycform(imageDouble,colorForm);
    imageDouble = imageDouble(:,:,1:channelCount); 
end
warning on

GaussianMask = torr_gauss_mask(windowSize, windowSize*GaussianVariance);
segmentCount = max(max(pixelLabels));
segments = cell(segmentCount,2);
sampleCount = floor(imageX*imageY);
sampleDimension = windowSize*windowSize*imageDepth;
X = zeros(sampleDimension,sampleCount);
maxMagnitude = 0;
% Construct sample vectors
for sampleIndex=1:sampleCount
    
    offsetY = floor((sampleIndex-1)/imageX)+1;
    offsetX = sampleIndex-(offsetY-1)*imageX;

    % (xIndex, yIndex) is inside of a region
    sampleWindow = imageDouble(offsetX:offsetX+2*halfWindowSize, offsetY:offsetY+2*halfWindowSize,:);
    for channelIndex=1:channelCount
        sampleWindow(:,:,channelIndex) = conv2(sampleWindow(:,:,channelIndex), GaussianMask, 'same');
    end
    X(:,sampleIndex) = reshape(sampleWindow,sampleDimension,1);
    
    % test if it is a boundary pixel
    windowX1=max(offsetX-halfWindowSize,1);
    windowX2=min(offsetX+halfWindowSize,imageX);
    windowY1=max(offsetY-halfWindowSize,1);
    windowY2=min(offsetY+halfWindowSize,imageY);
    labelWindow = pixelLabels(windowX1:windowX2, windowY1:windowY2);
    if isempty(find(labelWindow~=labelWindow(1)))
        % interior pixel
        XLabels(sampleIndex) = pixelLabels(sampleIndex)*2-1;
    else
        % boundary pixel
        XLabels(sampleIndex) = pixelLabels(sampleIndex)*2;
    end

    XNorm = norm(X(:,sampleIndex));
    if XNorm>maxMagnitude
        maxMagnitude = XNorm;
    end
end

% Normalize X
X = X/maxMagnitude;

% downsample
[U,S,V] = svds(X,downsampleDimension);
X = S*V.';

% Collect samples
for segmentIndex=1:segmentCount
    % interior samples
    segmentLabel = segmentIndex*2-1;
    segments{segmentIndex,1}=X(:,find(XLabels==segmentLabel));
    
    % boundary samples
    segmentLabel = segmentIndex*2;
    segments{segmentIndex,2}=X(:,find(XLabels==segmentLabel));
end